from io import BytesIO

from airflow.providers.sftp.hooks.sftp import SFTPHook

from airflow_tools.filesystems.filesystem_protocol import FilesystemProtocol


class SFTPFilesystem(FilesystemProtocol):
    def __init__(self, hook: SFTPHook):
        self.hook = hook

    def read(self, path: str) -> bytes:
        conn = self.hook.get_conn()
        out = BytesIO()
        conn.getfo(remotepath=path, fl=out)
        out.seek(0)
        return out.getvalue()

    def write(self, data: str | bytes | BytesIO, path: str):
        conn = self.hook.get_conn()
        if isinstance(data, str):
            data = data.encode()
        if isinstance(data, bytes):
            data = BytesIO(data)
        conn.putfo(fl=data, remotepath=path, confirm=True)

    def delete_file(self, path: str):
        self.hook.delete_file(path)

    def create_prefix(self, prefix: str):
        self.hook.create_directory(prefix)

    def delete_prefix(self, prefix: str):
        for file in self.hook.list_directory(prefix):
            self.hook.delete_file(prefix.rstrip('/') + '/' + file)
        self.hook.delete_directory(prefix)

    def check_file(self, path: str) -> bool:
        return self.hook.isfile(path)

    def check_prefix(self, prefix: str) -> bool:
        return self.hook.isdir(prefix)

    def list_files(self, prefix: str) -> list[str]:
        return self.hook.list_directory(prefix)



from functools import cached_property

from airflow.hooks.base import BaseHook
from airflow.models import Connection
from airflow.providers.common.sql.hooks.sql import DbApiHook
from databricks import sql
from databricks.sdk import WorkspaceClient
from databricks.sdk.core import Config, oauth_service_principal


class AzureDatabricksSqlHook(DbApiHook):
    """
    Requires defined connection with this structure:
        conn_type: azure_databricks_sql
        host: <hostname>
        login: <service_principal_id>
        password: <service_principal_secret>
        extra: {"http_path": "<http_path>", "catalog": "<catalog>", "schema": "<schema>"}
    """

    conn_name_attr = "azure_databricks_sql_conn_id"
    default_conn_name = "azure_databricks_sql_default"
    conn_type = "azure_databricks_sql"
    hook_name = "Azure Databricks SQL"

    def __init__(
        self,
        azure_databricks_sql_conn_id: str = default_conn_name,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.azure_databricks_sql_conn_id = azure_databricks_sql_conn_id
        self._sql_conn = None

    @cached_property
    def azure_databricks_sql_conn(self) -> Connection:
        return self.get_connection(self.azure_databricks_sql_conn_id)

    def _get_credentials(self):
        config = Config(
            host=f"https://{self.azure_databricks_sql_conn.host}",
            client_id=self.azure_databricks_sql_conn.login,
            client_secret=self.azure_databricks_sql_conn.password,
        )
        return oauth_service_principal(config)

    def get_conn(self) -> Connection:
        if not self._sql_conn:
            self._sql_conn = sql.connect(
                server_hostname=self.azure_databricks_sql_conn.host,
                http_path=self.azure_databricks_sql_conn.extra_dejson.get("http_path"),
                credentials_provider=self._get_credentials,
                catalog=self.azure_databricks_sql_conn.extra_dejson.get("catalog"),
                schema=self.azure_databricks_sql_conn.extra_dejson.get("schema"),
            )
        return self._sql_conn


class AzureDatabricksVolumeHook(BaseHook):
    conn_name_attr = "azure_databricks_volume_conn_id"
    default_conn_name = "azure_databricks_volume_default"
    conn_type = "azure_databricks_volume"
    hook_name = "Azure Databricks Volume"

    def __init__(
        self,
        azure_databricks_volume_conn_id: str = default_conn_name,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.azure_databricks_volume_conn_id = azure_databricks_volume_conn_id
        self.w = None

    @cached_property
    def azure_databricks_volume_conn(self) -> Connection:
        return self.get_connection(self.azure_databricks_volume_conn_id)

    def _get_config(self):
        return Config(
            host=f"https://{self.azure_databricks_volume_conn.host}",
            client_id=self.azure_databricks_volume_conn.login,
            client_secret=self.azure_databricks_volume_conn.password,
        )

    def _get_credentials(self):
        return oauth_service_principal(self._get_config())

    def get_conn(self) -> Connection:
        if not self.w:
            self.w = WorkspaceClient(
                host=self.azure_databricks_volume_conn.host, config=self._get_config()
            )
        return self.w



import json
from io import BytesIO, StringIO
from typing import TYPE_CHECKING, Any, Callable, Literal, Optional, Protocol

import jmespath
import pandas as pd
from airflow.hooks.base import BaseHook
from airflow.models import BaseOperator

try:
    from airflow.providers.http.operators.http import HttpOperator
except ImportError:
    from airflow.providers.http.operators.http import SimpleHttpOperator as HttpOperator

from airflow.utils.context import Context
from airflow.utils.helpers import merge_dicts
from requests import Response

from airflow_tools.compression_utils import CompressionOptions, compress
from airflow_tools.exceptions import ApiResponseTypeError
from airflow_tools.filesystems.filesystem_factory import FilesystemFactory

if TYPE_CHECKING:
    from requests.auth import AuthBase

SaveFormat = Literal['jsonl']


class Transformation(Protocol):
    def __call__(self, data: bytes, **kwargs) -> bytes:
        ...


class HttpBatchOperator(HttpOperator):
    def execute(
        self, context: Context, use_new_data_parameters_on_pagination=False
    ) -> Any:
        self.log.info("Calling HTTP method")

        response = self.hook.run(
            self.endpoint, self.data, self.headers, self.extra_options
        )
        yield self.process_response(context=context, response=response)
        for response in self.paginate_sync(
            response=response,
            use_new_data_parameters_on_pagination=use_new_data_parameters_on_pagination,
        ):
            yield self.process_response(context=context, response=response)

    def paginate_sync(
        self, response: Response, use_new_data_parameters_on_pagination=False
    ) -> Response | list[Response]:
        if not self.pagination_function:
            return None

        while True:
            next_page_params = self.pagination_function(response)
            if not next_page_params:
                break
            response = self.hook.run(
                **self._merge_next_page_parameters(
                    next_page_params, use_new_data_parameters_on_pagination
                )
            )
            yield response
        return None

    def _merge_next_page_parameters(
        self, next_page_params: dict, use_new_data_parameters_on_pagination=False
    ) -> dict:
        """Merge initial request parameters with next page parameters.

        Merge initial requests parameters with the ones for the next page, generated by
        the pagination function. Items in the 'next_page_params' overrides those defined
        in the previous request.

        :param next_page_params: A dictionary containing the parameters for the next page.
        :return: A dictionary containing the merged parameters.
        """
        data: str | dict | None = None  # makes mypy happy
        next_page_data_param = next_page_params.get("data")
        if use_new_data_parameters_on_pagination and isinstance(
            next_page_data_param, dict
        ):
            data = next_page_data_param
        elif isinstance(self.data, dict) and isinstance(next_page_data_param, dict):
            data = merge_dicts(self.data, next_page_data_param)
        else:
            data = next_page_data_param or self.data

        self.log.info(
            f"Calling HTTP method with endpoint: {next_page_params.get('endpoint') or self.endpoint}"
        )
        self.log.info(f"Calling HTTP method with data: {data}")

        return dict(
            endpoint=next_page_params.get("endpoint") or self.endpoint,
            data=data,
            headers=merge_dicts(self.headers, next_page_params.get("headers", {})),
            extra_options=merge_dicts(
                self.extra_options, next_page_params.get("extra_options", {})
            ),
        )


class HttpToFilesystem(BaseOperator):
    template_fields = list(HttpOperator.template_fields) + [
        'filesystem_path',
        'filesystem_conn_id',
        'jmespath_expression',
        'save_format',
        'source_format',
    ]
    template_fields_renderers = HttpOperator.template_fields_renderers

    json_response_source_format = ['json', 'jsonl']
    binary_response_source_format = ['parquet']

    def __init__(
        self,
        http_conn_id: str,
        filesystem_conn_id: str,
        filesystem_path: str,
        save_format: SaveFormat = 'jsonl',
        source_format: SaveFormat = None,
        compression: CompressionOptions = None,
        endpoint: str | None = None,
        method: str = "POST",
        data: Any = None,
        headers: dict[str, str] | None = None,
        auth_type: type['AuthBase'] | None = None,
        jmespath_expression: str | None = None,
        pagination_function: Callable | None = None,
        use_new_data_parameters_on_pagination: bool = False,
        create_file_on_success: str | None = None,
        data_transformation: Optional[Transformation] = None,
        data_transformation_kwargs: dict[str, Any] | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.http_conn_id = http_conn_id
        self.filesystem_conn_id = filesystem_conn_id
        self.filesystem_path = filesystem_path
        self.compression = compression
        self.endpoint = endpoint
        self.method = method
        self.data = data
        self.headers = headers
        self.auth_type = auth_type
        self.jmespath_expression = jmespath_expression
        self.pagination_function = pagination_function
        self.use_new_data_parameters_on_pagination = (
            use_new_data_parameters_on_pagination
        )
        self.create_file_on_success = create_file_on_success
        self.data_transformation = data_transformation
        self.data_transformation_kwargs = data_transformation_kwargs

        self.save_format = save_format
        self.source_format = source_format if source_format else save_format
        self.kwargs = kwargs

        if (
            self.save_format in self.binary_response_source_format
            and self.compression is not None
        ):
            raise ValueError(
                f'Compression is not supported for binary response save formats: {self.binary_response_source_format}'
            )

        if self.data_transformation and not callable(self.data_transformation):
            raise ValueError('data_transformation must be a callable')

        if self.data_transformation is None and self.source_format != self.save_format:
            raise ValueError(
                'data_transformation must be provided if source_format is different from save_format'
            )
        if self.data_transformation_kwargs and self.data_transformation is None:
            raise ValueError(
                'data_transformation must be provided if data_transformation_kwargs is provided'
            )

    def execute(self, context: 'Context') -> Any:
        http_batch_operator = HttpBatchOperator(
            task_id='http-operator',
            http_conn_id=self.http_conn_id,
            endpoint=self.endpoint,
            method=self.method,
            data=self.data,
            headers=self.headers,
            auth_type=self.auth_type,
            response_filter=self._response_filter,
            pagination_function=self.pagination_function,
        )
        for i, data in enumerate(
            http_batch_operator.execute(
                context,
                use_new_data_parameters_on_pagination=self.use_new_data_parameters_on_pagination,
            ),
            start=1,
        ):
            filesystem_protocol = FilesystemFactory.get_data_lake_filesystem(
                connection=BaseHook.get_connection(self.filesystem_conn_id),
            )

            file_path = self.filesystem_path.rstrip('/') + '/' + self._file_name(i)

            filesystem_protocol.write(data, file_path)

            if self.create_file_on_success is not None and isinstance(
                self.create_file_on_success, str
            ):
                success_file_path = (
                    self.filesystem_path.rstrip('/') + '/' + self.create_file_on_success
                )
                filesystem_protocol.write(BytesIO(), success_file_path)

    def _file_name(self, n_part) -> str:
        file_name = f'part{n_part:04}.{self.save_format}'

        if self.compression:
            file_name += f'.{self.compression}'
        return file_name

    def _response_filter(self, response) -> BytesIO:
        if (
            self.jmespath_expression
            and self.source_format in self.json_response_source_format
        ):
            self.data = jmespath.search(self.jmespath_expression, response.json())

        elif (
            self.jmespath_expression
            and self.source_format not in self.json_response_source_format
        ):
            raise ApiResponseTypeError(
                'JMESPath expression is only supported for json and jsonl save formats'
            )
        elif self.source_format in self.json_response_source_format:
            self.data = response.json()

        elif self.source_format in self.binary_response_source_format:
            self.data = response.content
        else:
            self.data = response.text

        # Check if we have a custom data transformation
        if self.data_transformation and self.data_transformation_kwargs:
            return self.data_transformation(self.data, self.data_transformation_kwargs)
        elif self.data_transformation:
            return self.data_transformation(self.data)

        # If we don't have a custom data transformation, use the default one based on the source_format

        match self.source_format:
            case 'json':
                return json_to_binary(self.data, self.compression)

            case 'jsonl':
                if not isinstance(self.data, list):
                    raise ApiResponseTypeError(
                        'Expected response can\'t be transformed to jsonl. It is not  list[dict]'
                    )
                return list_to_jsonl(self.data, self.compression)

            case 'xml':
                return xml_to_binary(self.data, self.compression)

            case 'parquet':
                return self.data

            case 'csv':
                return csv_to_binary(self.data, self.compression)

            case _:
                raise NotImplementedError(
                    f'Unknown source_format/save_format: {self.source_format}'
                )


def list_to_jsonl(data: list[dict], compression: 'CompressionOptions') -> BytesIO:
    out = StringIO()
    df = pd.DataFrame(data)
    df.to_json(out, orient='records', lines=True, compression=compression)
    out.seek(0)
    return BytesIO(out.getvalue().encode())


def json_to_binary(data: dict, compression: 'CompressionOptions') -> BytesIO:
    json_string = json.dumps(data).encode()
    compressed_json = compress(compression, json_string)
    out = BytesIO(compressed_json)
    return out


def csv_to_binary(data: str, compression: 'CompressionOptions') -> BytesIO:
    csv_df = pd.read_csv(StringIO(data), sep=',')
    out = BytesIO()
    csv_df.to_csv(out, compression=compression, index=False)
    out.seek(0)
    return out


def xml_to_binary(data: str, compression: 'CompressionOptions') -> BytesIO:
    compressed_xml = compress(compression, data.encode())
    out = BytesIO(compressed_xml)
    return out



import io
import logging
import sys
import typing

import pandas as pd
from airflow.hooks.base import BaseHook
from airflow.models import BaseOperator
from sqlalchemy import create_engine

from airflow_tools.filesystems.filesystem_factory import FilesystemFactory

if sys.version_info < (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

logger = logging.getLogger(__name__)


class FilesystemToDatabaseOperator(BaseOperator):
    """
    This operator will copy a file from a filesystem to a database table.
    """

    template_fields = (
        'db_table',
        'db_schema',
        'filesystem_conn_id',
        'database_conn_id',
        'filesystem_path',
        'metadata',
    )

    def __init__(
        self,
        filesystem_conn_id: str,
        database_conn_id: str,
        filesystem_path: str,
        db_table: str,
        db_schema: typing.Optional[str] = None,
        source_format: typing.Optional[Literal['csv', 'json', 'parquet']] = 'csv',
        source_format_options: typing.Optional[typing.Dict] = None,
        table_aggregation_type: typing.Optional[
            Literal['append', 'fail', 'replace']
        ] = 'append',
        metadata: typing.Optional[typing.Dict[str, str]] = None,
        include_source_path: bool = True,
        *args,
        **kwargs,
    ) -> None:

        super().__init__(*args, **kwargs)

        self.filesystem_conn_id = filesystem_conn_id
        self.database_conn_id = database_conn_id
        self.filesystem_path = filesystem_path
        self.db_table = db_table
        self.db_schema = db_schema
        self.source_format = source_format
        self.source_format_options = source_format_options
        self.table_aggregation_type = table_aggregation_type
        self.metadata = metadata or {'_DS': '{{ ds }}'}
        self.include_source_path = include_source_path

    def execute(self, context):
        logger.info(f'Create connection for filesystem ({self.filesystem_conn_id})')
        filesystem = FilesystemFactory.get_data_lake_filesystem(
            connection=BaseHook.get_connection(self.filesystem_conn_id),
        )

        logger.info(f'Create SQLAlchemy engine with connection_id {self.database_conn_id}')
        engine = create_engine(
            BaseHook.get_connection(self.database_conn_id).get_hook().get_uri()
        )
        
        for blob_path in filesystem.list_files(prefix=self.filesystem_path):

            if not blob_path.endswith(
                (f'.{self.source_format}', f'.{self.source_format}.gz')
            ):
                logger.warning(f'Blob {blob_path} is not in the right format. Skipping...')
                continue

            logger.info(f'Read file {blob_path} and convert to pandas')
            raw_content = io.BytesIO(filesystem.read(blob_path))

            df = self.raw_content_to_pandas(path_or_buf=raw_content)

            for key, value in self.metadata.items():
                df[key] = value
                
                if key.startswith('_'):
                    # _* fields are treated as metadata and we try to convert them in datetimes
                    df[key] = self._convert_to_datetime(df[key])
                    
            if self.include_source_path:
                df['_LOADED_FROM'] = blob_path
                df['_LOADED_FROM'] = df['_LOADED_FROM'].astype('string')

            df.to_sql(
                name=self.db_table,
                schema=self.db_schema,
                con=engine,
                if_exists=self.table_aggregation_type,
                index=False,
            )
            
    @staticmethod
    def _convert_to_datetime(value):
        try:
            return pd.to_datetime(value)
        except ValueError:
            return value

    def raw_content_to_pandas(self, path_or_buf: typing.Union[str, bytes, io.StringIO]):
        options = self.source_format_options or {}

        match self.source_format:
            case 'csv':
                return pd.read_csv(path_or_buf, **options)
            case 'json':
                return pd.read_json(path_or_buf, **options)
            case 'parquet':
                return pd.read_parquet(path_or_buf, **options)
            case _:
                raise ValueError(f'Unknown source format {self.source_format}')



import gzip
import zipfile
from io import BytesIO
from typing import Literal, Union

DEFAULT_ZIP_FILENAME = 'file.zip'
CompressionOptions = Union[Literal['infer', 'gzip', 'bz2', 'zip', 'xz', 'zstd'], None]


def gzip_data(data: bytes) -> bytes:
    out = BytesIO()
    with BytesIO() as out:
        with gzip.GzipFile(fileobj=out, mode="wb") as f:
            f.write(data)
        out.seek(0)
        return out.getvalue()


def zip_data(data: bytes, filename: str = DEFAULT_ZIP_FILENAME):
    with BytesIO() as out:
        with zipfile.ZipFile(out, 'w', zipfile.ZIP_DEFLATED) as zip_file:
            zip_file.writestr(filename, data)

        out.seek(0)
        return out.getvalue()


def compress(compression: 'CompressionOptions', data: bytes) -> bytes:
    if compression is None:
        return data
    if compression == 'gzip':
        return gzip_data(data)
    if compression == 'zip':
        return zip_data(data)
    raise NotImplementedError(f'Unknown compression: {compression}')


def ungzip_data(compressed_data: bytes) -> bytes:
    with BytesIO(compressed_data) as input_stream:
        with gzip.GzipFile(fileobj=input_stream, mode="rb") as f:
            return f.read()


def unzip_data(compressed_data: bytes, filename: str = DEFAULT_ZIP_FILENAME) -> bytes:
    with BytesIO(compressed_data) as input_stream:
        with zipfile.ZipFile(input_stream) as zip_file:
            with zip_file.open(filename) as file:
                return file.read()


def decompress(compression: 'CompressionOptions', data: bytes) -> bytes:
    if compression is None:
        return data
    if compression == 'gzip':
        return ungzip_data(data)
    if compression == 'zip':
        return unzip_data(data)
    raise NotImplementedError(f'Unknown compression: {compression}')



from io import BytesIO
from typing import TYPE_CHECKING, Optional, Protocol

from airflow.hooks.base import BaseHook
from airflow.models import BaseOperator
from airflow.providers.common.sql.hooks.sql import DbApiHook

from airflow_tools.filesystems.filesystem_factory import FilesystemFactory

if TYPE_CHECKING:
    from airflow.utils.context import Context

import logging

logger = logging.getLogger(__file__)


class Transformation(Protocol):
    def __call__(self, data: bytes, filename: str, context: dict) -> bytes:
        ...


class SQLToFilesystem(BaseOperator):
    """
    Copies data from a SQL query to a filesystem.
    """

    template_fields = (
        'sql',
        'destination_path',
    )

    def __init__(
        self,
        source_sql_conn_id: str,
        destination_fs_conn_id: str,
        sql: str,
        destination_path: str,
        batch_size: int | None = None,
        create_file_on_success: str | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.source_sql_conn_id = source_sql_conn_id
        self.destination_fs_conn_id = destination_fs_conn_id
        self.sql = sql
        self.destination_path = destination_path
        self.batch_size = batch_size
        self.create_file_on_success = create_file_on_success
        self.files = []

    def execute(self, context):
        source_sql_hook: DbApiHook = BaseHook.get_connection(
            self.source_sql_conn_id
        ).get_hook()

        destination_fs_hook = FilesystemFactory.get_data_lake_filesystem(
            connection=BaseHook.get_connection(self.destination_fs_conn_id),
        )

        self.files.clear()
        dfs = (
            [source_sql_hook.get_pandas_df(sql=self.sql)]
            if not self.batch_size
            else source_sql_hook.get_pandas_df_by_chunks(
                sql=self.sql, chunksize=self.batch_size
            )
        )
        for i, df in enumerate(dfs, start=1):
            full_file_path = f"{self.destination_path.rstrip('/')}/part{i:04}.parquet"
            destination_fs_hook.write(df.to_parquet(index=False), full_file_path)
            self.files.append(full_file_path)

        if self.create_file_on_success is not None and isinstance(self.create_file_on_success, str):
            success_file_path = f"{self.destination_path.rstrip('/')}/{self.create_file_on_success}"
            destination_fs_hook.write(BytesIO(), success_file_path)


class FilesystemToFilesystem(BaseOperator):
    """
    Copies a file from a filesystem to another filesystem.
    """

    template_fields = ('source_path', 'destination_path')

    def __init__(
        self,
        source_fs_conn_id: str,
        destination_fs_conn_id: str,
        source_path: str,
        destination_path: str,
        data_transformation: Optional[Transformation] = None,
        create_file_on_success: str | None = None,
        *args,
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.source_fs_conn_id = source_fs_conn_id
        self.destination_fs_conn_id = destination_fs_conn_id
        self.source_path = source_path
        self.destination_path = destination_path
        self.data_transformation = data_transformation
        self.create_file_on_success = create_file_on_success

    def execute(self, context):
        source_fs_hook = FilesystemFactory.get_data_lake_filesystem(
            connection=BaseHook.get_connection(self.source_fs_conn_id),
        )
        destination_fs_hook = FilesystemFactory.get_data_lake_filesystem(
            connection=BaseHook.get_connection(self.destination_fs_conn_id),
        )

        files = (
            source_fs_hook.list_files(self.source_path)
            if source_fs_hook.check_prefix(self.source_path)
            else [self.source_path]
        )
        for file_path in files:
            logger.info(f'Trying to copy {file_path} to {self.destination_path}')
            file_name = file_path.split('/')[-1]
            data = source_fs_hook.read(file_path)
            if self.data_transformation:
                data = self.data_transformation(data, file_name, context)
            full_destination_path = (
                self.destination_path + file_name
                if self.destination_path.endswith('/')
                else self.destination_path
            )
            destination_fs_hook.write(data, full_destination_path)

        if self.create_file_on_success is not None and isinstance(self.create_file_on_success, str):
            success_file_path = '/'.join(self.destination_path.split('/')[:-1]) + '/' + self.create_file_on_success
            destination_fs_hook.write(BytesIO(), success_file_path)


class FilesystemDeleteOperator(BaseOperator):
    template_fields = ('filesystem_path',)

    def __init__(self, filesystem_conn_id: str, filesystem_path: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.filesystem_conn_id = filesystem_conn_id
        self.filesystem_path = filesystem_path

    def execute(self, context: 'Context'):
        filesystem_protocol = FilesystemFactory.get_data_lake_filesystem(
            connection=BaseHook.get_connection(self.filesystem_conn_id),
        )
        logger.info(f'Trying to delete: {self.filesystem_path}')
        filesystem_protocol.delete_prefix(self.filesystem_path)


class FilesystemCheckOperator(BaseOperator):
    template_fields = ('filesystem_path',)

    def __init__(
        self,
        filesystem_conn_id: str,
        filesystem_path: str,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.filesystem_conn_id = filesystem_conn_id
        self.filesystem_path = filesystem_path

    def execute(self, context: 'Context'):
        filesystem_protocol = FilesystemFactory.get_data_lake_filesystem(
            connection=BaseHook.get_connection(self.filesystem_conn_id),
        )

        logger.info(f'Trying to check: {self.filesystem_path}')
        return (
            filesystem_protocol.check_prefix(self.filesystem_path)
            if self.filesystem_path.endswith('/')
            else filesystem_protocol.check_file(self.filesystem_path)
        )



import sys
import logging

from airflow.models import BaseOperator
from airflow.hooks.base import BaseHook

from deltalake import write_deltalake
from duckdb_provider.hooks.duckdb_hook import DuckDBHook

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

logger = logging.getLogger(__name__)


class DuckdbToDeltalakeOperator(BaseOperator):
    template_fields = ('source_query', 'table_path')

    def __init__(
        self,
        duckdb_conn_id: str,
        delta_lake_conn_id: str,
        source_query: str,
        table_path: str,
        extensions: list[str] = None,
        rows_per_batch: int = 1000000,
        write_mode: Literal["error", "append", "overwrite", "ignore"] = "append",
        *args,
        **kwargs
    ) -> None:
        super().__init__(*args, **kwargs)
        self.duckdb_conn_id = duckdb_conn_id
        self.delta_lake_conn_id = delta_lake_conn_id
        self.source_query = source_query
        self.table_path = table_path
        self.extensions = extensions
        self.rows_per_batch = rows_per_batch
        self.write_mode = write_mode

    def execute(self, context):
        logger.info("Create conn for duckdb")
        duckdb_conn = DuckDBHook.get_hook(self.duckdb_conn_id).get_conn()

        if self.extensions:
            logger.info("Install and load extensions")
            for extension in self.extensions:
                duckdb_conn.install_extension(extension)
                duckdb_conn.load_extension(extension)

        logger.info("Create conn for delta lake and populate storage_options")
        delta_lake_conn = BaseHook.get_connection(self.delta_lake_conn_id)
        connection_string = delta_lake_conn.extra_dejson.get("connection_string")

        logger.info("Create secret for duckdb")
        duckdb_conn.execute(f"""
            CREATE OR REPLACE SECRET delta_lake_secret (
                TYPE AZURE,
                CONNECTION_STRING '{connection_string}'
            );""")

        logger.info("Write data to delta lake")
        write_deltalake(self.table_path,
                        duckdb_conn.execute(self.source_query).fetch_record_batch(rows_per_batch=self.rows_per_batch),
                        storage_options=self.parse_storage_options(connection_string),
                        mode=self.write_mode)

    @staticmethod
    def parse_storage_options(connection_string: str) -> dict[str, str]:
        storage_options = {}
        for part in connection_string.split(';'):
            if part.startswith('AccountName='):
                storage_options['account_name'] = part.replace('AccountName=', '').strip()
            if part.startswith('AccountKey='):
                storage_options['account_key'] = part.replace('AccountKey=', '').strip()
        return storage_options



from typing import TYPE_CHECKING

from airflow.hooks.base import BaseHook
from airflow.models import BaseOperator

from airflow_tools.data_lake_facade import DataLakeFacade

if TYPE_CHECKING:
    from airflow.utils.context import Context

import logging

logger = logging.getLogger(__file__)


class DataLakeDeleteOperator(BaseOperator):
    template_fields = ('data_lake_path',)

    def __init__(self, data_lake_conn_id: str, data_lake_path: str, *args, **kwargs):
        logger.warning('DataLakeDeleteOperator is deprecated. Use FilesystemDeleteOperator instead.')
        super().__init__(*args, **kwargs)
        self.data_lake_conn_id = data_lake_conn_id
        self.data_lake_path = data_lake_path

    def execute(self, context: 'Context'):
        data_lake_conn = BaseHook.get_connection(self.data_lake_conn_id)
        data_lake_facade = DataLakeFacade(
            conn=data_lake_conn.get_hook(),
        )
        data_lake_facade.delete_prefix(self.data_lake_path)


class DataLakeCheckOperator(BaseOperator):
    template_fields = ('data_lake_path',)

    def __init__(
        self,
        data_lake_conn_id: str,
        data_lake_path: str,
        check_specific_filename: None | str = None,
        *args,
        **kwargs,
    ):
        logger.warning('DataLakeCheckOperator is deprecated. Use FilesytemCheckOperator instead.')
        super().__init__(*args, **kwargs)
        self.data_lake_conn_id = data_lake_conn_id
        self.data_lake_path = data_lake_path
        self.check_specific_filename = check_specific_filename

    def execute(self, context: 'Context'):
        data_lake_conn = BaseHook.get_connection(self.data_lake_conn_id)
        data_lake_facade = DataLakeFacade(
            conn=data_lake_conn.get_hook(),
        )
        logger.info(f'Checking {self.data_lake_path}')
        prefix_flag = data_lake_facade.check_prefix(self.data_lake_path)
        logger.info(f'Prefix flag: {prefix_flag}')
        if self.check_specific_filename:
            logger.info(f'Checking {self.check_specific_filename}')
            specific_file_flag = data_lake_facade.check_prefix(
                f'{self.data_lake_path}{self.check_specific_filename}'
            )
            logger.info(f'Specific file flag: {specific_file_flag}')
            return prefix_flag and specific_file_flag
        return prefix_flag



from typing import Any, Dict

from airflow_tools._version import __version__


def get_provider_info() -> Dict[str, Any]:
    """Return provider metadata to Airflow"""
    return {
        "package-name": "airflow-tools",
        "name": "DKL Airflow Tools",
        "description": "Apache Airflow Providers containing Operators, Sensors and tools for ELT",
        "versions": __version__,
        # Optional.
        "connection-types": [
            {
                "connection-type": "azure_databricks_sql",
                "hook-class-name": "airflow_tools.providers.azure.hooks.azure_databricks.AzureDatabricksSqlHook",
            },
            {
                "connection-type": "azure_databricks_volume",
                "hook-class-name": "airflow_tools.providers.azure.hooks.azure_databricks.AzureDatabricksVolumeHook",
            },
            {
                "connection-type": "azure_file_share_sp",
                "hook-class-name": "airflow_tools.providers.azure.hooks.azure_file_share.AzureFileShareServicePrincipalHook",
            },
        ],
        "extra-links": [],
    }
